/*
  (c) FSF 2002. 
*/
   

#ifndef PRIVATE_GC
#define PRIVATE_GC

#include  "_scm.h"

/* {heap tuning parameters}
 *
 * These are parameters for controlling memory allocation.  The heap
 * is the area out of which scm_cons, and object headers are allocated.
 *
 * Each heap cell is 8 bytes on a 32 bit machine and 16 bytes on a
 * 64 bit machine.  The units of the _SIZE parameters are bytes.
 * Cons pairs and object headers occupy one heap cell.
 *
 * SCM_INIT_HEAP_SIZE is the initial size of heap.  If this much heap is
 * allocated initially the heap will grow by half its current size
 * each subsequent time more heap is needed.
 *
 * If SCM_INIT_HEAP_SIZE heap cannot be allocated initially, SCM_HEAP_SEG_SIZE
 * will be used, and the heap will grow by SCM_HEAP_SEG_SIZE when more
 * heap is needed.  SCM_HEAP_SEG_SIZE must fit into type size_t.  This code
 * is in scm_init_storage() and alloc_some_heap() in sys.c
 *
 * If SCM_INIT_HEAP_SIZE can be allocated initially, the heap will grow by
 * SCM_EXPHEAP(scm_heap_size) when more heap is needed.
 *
 * SCM_MIN_HEAP_SEG_SIZE is minimum size of heap to accept when more heap
 * is needed.
 *
 * INIT_MALLOC_LIMIT is the initial amount of malloc usage which will
 * trigger a GC.
 *
 * SCM_MTRIGGER_HYSTERESIS is the amount of malloc storage that must
 * be reclaimed by a GC triggered by a malloc. If less than this is
 * reclaimed, the trigger threshold is raised. [I don't know what a
 * good value is. I arbitrarily chose 1/10 of the INIT_MALLOC_LIMIT to
 * work around a oscillation that caused almost constant GC.]
 */

/*
 * Heap size 45000 and 40% min yield gives quick startup and no extra
 * heap allocation.  Having higher values on min yield may lead to
 * large heaps, especially if code behaviour is varying its
 * maximum consumption between different freelists.
 */


#define scm_default_init_heap_size_1  256*1024
#define scm_default_min_yield_1 40
#define scm_default_init_heap_size_2 32*1024

/* The following value may seem large, but note that if we get to GC at
 * all, this means that we have a numerically intensive application
 */
#define  scm_default_min_yield_2 40

#define scm_default_max_segment_size  2097000L /* a little less (adm) than 2 Mb */

#define SCM_MIN_HEAP_SEG_SIZE (8 * SCM_GC_SIZEOF_CARD)
#define SCM_HEAP_SEG_SIZE (16384L * sizeof (scm_t_cell))

/* Make heap grow with factor 1.5 */
#define SCM_EXPHEAP(scm_heap_size) (scm_heap_size / 2)
#define SCM_INIT_MALLOC_LIMIT 100000
#define SCM_MTRIGGER_HYSTERESIS (SCM_INIT_MALLOC_LIMIT/10)

/* CELL_UP and CELL_DN are used by scm_init_heap_seg to find (scm_t_cell * span)
   aligned inner bounds for allocated storage */


# ifdef _UNICOS
#  define CELL_UP(p, span) (SCM_CELLPTR)(~(span) & ((long)(p)+(span)))
#  define CELL_DN(p, span) (SCM_CELLPTR)(~(span) & (long)(p))
# else
#  define CELL_UP(p, span) (SCM_CELLPTR)(~(sizeof(scm_t_cell)*(span)-1L) & ((long)(p)+sizeof(scm_t_cell)*(span)-1L))
#  define CELL_DN(p, span) (SCM_CELLPTR)(~(sizeof(scm_t_cell)*(span)-1L) & (long)(p))
# endif				/* UNICOS */


#define DOUBLECELL_ALIGNED_P(x)  (((2 * sizeof (scm_t_cell) - 1) & SCM_UNPACK (x)) == 0)

// debugging.
typedef struct scm_t_list_cell_struct {
  scm_t_bits car;  
  struct scm_t_list_cell_struct * cdr;
} scm_t_list_cell;



int scm_getenv_int (const char *var, int def);


typedef enum { return_on_error, abort_on_error } policy_on_error;

/* gc-freelist*/

/*
  FREELIST:

  A struct holding GC statistics on a particular type of cells.
*/
typedef struct scm_t_freelist {

  /*
    heap segment where the last cell was allocated 
  */
  int heap_segment_idx;

  /* minimum yield on this list in order not to grow the heap
   */
  long min_yield;

  /* defines min_yield as percent of total heap size
   */
  int min_yield_fraction;
  
  /* number of cells per object on this list */
  int span;

  /* number of collected cells during last GC */
  unsigned long collected;

  /* number of collected cells during penultimate GC */
  unsigned long collected_1;

  /* total number of cells in heap segments
   * belonging to this list.
   */
  unsigned long heap_size;

} scm_t_freelist;


extern scm_t_freelist scm_master_freelist;
extern scm_t_freelist scm_master_freelist2;
extern unsigned long scm_gc_yield_1;

void scm_adjust_min_yield (scm_t_freelist *freelist);
void scm_gc_init_freelist (void);
void scm_initialize_freelist (scm_t_freelist *freelist,
			      int span,
			      int min_yield);

void scm_gc_sweep_freelist_reset (scm_t_freelist *freelist);
int scm_gc_grow_heap_p (scm_t_freelist * freelist);
     
#define SCM_HEAP_SIZE \
  (scm_master_freelist.heap_size + scm_master_freelist2.heap_size)


#define SCM_MAX(A, B) ((A) > (B) ? (A) : (B))
#define SCM_MIN(A, B) ((A) < (B) ? (A) : (B))

#define CELL_P(x)  (SCM_ITAG3 (x) == scm_tc3_cons)

/*
  gc-mark
 */


void scm_mark_all (void);
void scm_init_gc_mark (void);


/*
gc-segment:
*/




/*

 Cells are stored in a heap-segment: it is a contiguous chunk of
 memory, that associated with one freelist. 
*/

typedef struct scm_t_heap_seg_data
{
  /*
    {lower, upper} bounds of the segment

    The upper bound is also the start of the mark space.
  */
  scm_t_cell *bounds[2];

  /*
    If we ever decide to give it back, we could do it with this ptr.

    Note that giving back memory is not very useful; as long we don't
    touch a chunk of memory, the virtual memory system will keep it
    swapped out. We could simply forget about a block.

    (not that we do that, but anyway.) 
   */

  void* malloced;

  scm_t_cell * next_free_card;
  
  /* address of the head-of-freelist pointer for this segment's cells.
     All segments usually point to the same one, scm_freelist.  */
  scm_t_freelist *freelist;
  
  /* number of cells per object in this segment */
  int span;


  /*
    Is this the first time that the cells are accessed? 
   */
  int first_time;
  
} scm_t_heap_seg_data;



/*

  A table of segment records is kept that records the upper and
  lower extents of the segment;  this is used during the conservative
  phase of gc to identify probably gc roots (because they point
  into valid segments at reasonable offsets).

*/
extern scm_t_heap_seg_data ** scm_heap_segment_table;
extern size_t scm_heap_segment_table_size;


int scm_init_card_freelist (scm_t_cell * card, SCM *free_list,int);
int scm_sweep_card (scm_t_cell * card, SCM *free_list,int);
int scm_sweep_segment (scm_t_heap_seg_data *seg);
int scm_initialize_heap_segment_data (scm_t_heap_seg_data * segment, size_t requested);
int scm_segment_card_count (scm_t_heap_seg_data * seg);
int scm_segment_cell_count (scm_t_heap_seg_data * seg);

void scm_segment_clear_mark_space (scm_t_heap_seg_data *seg);
scm_t_heap_seg_data * scm_make_empty_heap_segment (scm_t_freelist*);
void scm_heap_segment (scm_t_heap_seg_data *h);
SCM scm_sweep_some_cards (scm_t_heap_seg_data *seg);

int scm_insert_segment (scm_t_heap_seg_data * seg);
long int scm_find_heap_segment_containing_object (SCM obj);
int scm_get_new_heap_segment (scm_t_freelist *, policy_on_error);
void scm_clear_mark_space (void);
void scm_sweep_segments (void);
SCM scm_sweep_some_segments (scm_t_freelist * fl);
void scm_reset_segments (void);

void scm_make_initial_segment (size_t init_heap_size, scm_t_freelist *freelist);
void scm_gc_init_segments (void);

/*
  gc-sweep
 */

extern long int scm_deprecated_memory_return;
void scm_gc_sweep (void);


     
#endif
